-- ***************************************************************************
-- 'Recursion' v0.1b.1 by dragoonreas, 2010/08/05
--
-- Features:
-- > A tree fractal algorithm to move asteroids in recursive orbits.
-- > Utility and feature functions for more modular code.
-- > A debug mode for troubleshooting.
--
-- Notes:
-- > As evident my the version number, this is by no means a final release.
-- > I hope to update this further over the course of the coming weeks.
--
-- Version History:
-- v0.1b.1	> BUGFIX: Stopped crash on startup when failing to create 
--				debugging file while not in debug mode.
-- v0.1b	> Initial beta release.
-- ***************************************************************************

function LevelSetup()
	
	-- Set Debugging Values
	DEBUG = false
	if DEBUG == true then
		-- Will output debug info to a text file in the Output folder, which should be on the same level as Eufloria.exe
		io.output("Output/Recursion Level Output.txt") -- TODO: Change this (and io.write() lines) to use the Complete I/O Model so error checking on the avaliability of the output file can be done
	end
	
	
	-- Set Global Values
	Globals.G.Asteroids=(0)
	Globals.Structures.FlowerProbability=(0) -- Flowers don't follow moving asteroids once planted, but still provide the tree with superseedling / mines (bug in v2.05)
	
	
	-- Set Fractal Values
	depth = 2
	falloff = 0.85
	controllers = 3
	
	phases = {}
	for i = 0, controllers - 1 do
		phases[i] = 360 / controllers * i - 360 / (2 * controllers)
	end
	
	amplitude = 2400 -- TODO: Put more distance between the asteroids
	diameter = amplitude / (2 * depth * controllers)
	period = math.pow(controllers, 2) * 60 -- TODO: Might need to slow everything down a bit
	
	roidCount = 0
	layerCounts = {}
	for i = 0, controllers do
		roidCount = roidCount + math.pow(controllers, i)
		layerCounts[i] = roidCount
	end
	
	
	-- Create Asteroids & Empires
	angles = {}
	for i = 0, roidCount - 1 do
		local currentDepth = GetDepth(i)
		local falloffMultiplier = math.pow(falloff, currentDepth)
		
		local a = AddAsteroid(0,0)
		a:SetRadius(diameter * falloffMultiplier) -- TODO: Increase the size difference between asteroids of different depths
		a.Attribs.Energy = falloffMultiplier
		a.Attribs.Strength = falloffMultiplier
		a.Attribs.Speed = falloffMultiplier
		a.MaxCoreHealth = DecToInt(100 * falloffMultiplier)
		a.SendDistance = 10 * a.Radius
		a.Moveable = False
		angles[i] = 0
		
		-- Debugging Stuff
		if DEBUG == true then
			local roidType = -1
			if falloffMultiplier < 1 then
				roidType = (i - layerCounts[currentDepth - 1]) % 3
			end
			
			a.Name = string.format("Asteroid %i:D%iT%i", i, currentDepth, roidType)
			a:Reveal(1)
			
			-- TODO: Find out why the SendDistance is so large for some of the asteroids despite them being reported as consistently sized (a global variable that needs to be set, or maybe even a bug?)
			io.write(string.format("roidID %i: SendDistance = %.3f\n", i, a.SendDistance))
		end
		
		if falloffMultiplier < 1 then
			local prevFalloffMultiplier = math.pow(falloff, currentDepth - 1)
			local roidType = (i - layerCounts[currentDepth - 1]) % 3

			if roidType == 0 then
				a.Attribs.Energy = prevFalloffMultiplier
			elseif roidType == 1 then
				a.Attribs.Strength = prevFalloffMultiplier
			else
				a.Attribs.Speed = prevFalloffMultiplier
			end
		end
		
		if i == 0 then
			centralRoid = a
			
			a.Owner = 1
			a:AddSeedlings(50)
			
			local s = a:AddDysonTree()
			s:LevelUp()
			s:LevelUp()
			s:LevelUp()
			
			s = a:AddDefenseTree()
			s:LevelUp()
			s:LevelUp()
			s:LevelUp()

			s = a:AddDefenseTree()
			s:LevelUp()
			s:LevelUp()
			s:LevelUp()
		elseif currentDepth == 0 then
			a.Owner = 1 + i
			a:AddSeedlings(50)
			
			local s = a:AddDysonTree()
			s:LevelUp()
			s:LevelUp()
			s:LevelUp()
			
			s = a:AddDefenseTree()
			s:LevelUp()
			s:LevelUp()
			s:LevelUp()

			s = a:AddDefenseTree()
			s:LevelUp()
			s:LevelUp()
			s:LevelUp()
		elseif currentDepth == 1 then
			a.Owner = 0
			a:AddSuperSeedlings(DecToInt(500 / (controllers * 2 + currentDepth)))
		end
	end

	MoveAsteroids()
end



function LevelLogic()
	while GameRunning() do
		
		-- Check if Player has Won
		for i = 2, controllers + 1 do
			if GetEmpire(i):GetNumOwnedAsteroids() > 0 then
				break
			end
			
			if i == controllers + 1 then
				if depth > 0 then
					if GetEmpire(0):GetNumOwnedAsteroids() > 0 then
						break
					end
				end
				
				Pause()
				MessageBox("You have won.")
				WaitDialogue()
				Unpause()
				
				Quit(true)
			end
		end
		
		
		-- Check if Player has Lost
		if GetEmpire(1):GetNumOwnedAsteroids() == 0 then
			Pause()
			MessageBox("You have lost.")
			WaitDialogue()
			Unpause()
			
			Quit(false)
		end
		
		-- Update Asteroids
		MoveAsteroids()
		
		
		-- Let the Game do its thing
		coroutine.yield()
	end
end



-- Functions not prototyped in the game, just defined for use in the lua script

-- Use a tree fractal algorithm to move the asteroids 
function MoveAsteroids()
	for i = 0, depth do
		if i == 0 then
			for j = 0, controllers - 1 do
				local roidID = j + layerCounts[i]
				local x = math.sin(((GetGameTime() % period) / period + phases[j] / 360) * 2 * math.pi) * amplitude
				local y = math.cos(((GetGameTime() % period) / period + phases[j] / 360) * 2 * math.pi) * amplitude
				
				GetAsteroid(roidID):MoveTo(x,y)
				angles[roidID] = math.atan2(x, y)
			end
		else
			local prevStartIndex = layerCounts[i - 1]
			local startIndex = layerCounts[i]
			local prevCount = startIndex - prevStartIndex
			
			for j = 0, prevCount - 1 do
				local orbitedRoidID = prevStartIndex + j
				local orbitedRoid = GetAsteroid(orbitedRoidID)
				
				for k = 0, controllers - 1 do
					local roidID = startIndex + j * controllers + k
					local controllerRoid = GetAsteroid(k + 1)
					local distance = math.sqrt(math.pow(controllerRoid.Position.X + centralRoid.Position.X, 2) + math.pow(controllerRoid.Position.Y - centralRoid.Position.Y, 2)) -- TODO: Swap out this line for the one below once it's working
					--local distance = controllerRoid:DistanceSquaredTo(centralRoid) -- TODO: Find out why this gives the error "Object reference not set to an instance of an object."
					local angle = angles[orbitedRoidID] + math.atan2(controllerRoid.Position.Y, controllerRoid.Position.X) - math.rad(90)
					local x = orbitedRoid.Position.X + (distance * math.pow(falloff, i)) * math.sin(angle)
					local y = orbitedRoid.Position.Y + (distance * math.pow(falloff, i)) * math.cos(angle)
					
					GetAsteroid(roidID):MoveTo(x,y)
					angles[roidID] = angle
				end
			end
		end
	end
end



-- Get the depth of an asteroid within the fractal using the asteroid id
function GetDepth(roidID)
	local currentDepth = 1
	
	for i = 0, controllers do
		if layerCounts[i] <= roidID then
			currentDepth = currentDepth + 1
		else
			currentDepth = currentDepth - 1
			break
		end
	end
	
	return currentDepth
end



-- Rounds a decimal / floating point number to a given precision
function DecimalPrecision(decimal, precision)
	return DecToInt(decimal * math.pow(10, precision)) / math.pow(10, precision)
end



-- Rounds a decimal / floating point number to the nearest integer
function DecToInt(decimal)
	if decimal - math.floor(decimal) < 0.5 then
		return math.floor(decimal)
	else
		return math.ceil(decimal)
	end
end
